// GPSNet.cpp
//
// Copyright (C) 2013 Mike McCauley
// $Id: GPSNet.cpp,v 1.1 2013/02/20 06:43:38 mikem Exp mikem $

#include "GPSNet.h"
GPSReport::GPSReport()
{
    magic = GPSNET_MAGIC_VERSION_1;
    date = 0;
    time = 0;
    latitude = 0;
    longitude = 0;
    altitude = 0;
    speed = 0;
    course = 0;
    flags = 0;
    status = 0;
    hops = 0;
}

boolean GPSReport::hasId(uint8_t* id)
{
    return memcmp(id, this->id, GPSNET_REPORT_ID_LEN) == 0;
}

boolean GPSReport::isIdenticalTo(GPSReport* that)
{
    return hasId(that->id)
	&& this->date      == that->date
	&& this->time      == that->time
	&& this->latitude  == that->latitude
	&& this->longitude == that->longitude
	&& this->altitude  == that->altitude
	&& this->speed     == that->speed
	&& this->course    == that->course
	&& this->status    == that->status
	&& this->flags    == that->flags
	&& this->hops      == that->hops
	;
}

boolean GPSReport::isCloseTo(GPSReport* that, uint32_t latlongError, uint32_t altitudeError)
{
    return (labs(this->latitude  - that->latitude)  < latlongError)
	&& (labs(this->longitude - that->longitude) < latlongError)
	&& (labs(this->altitude  - that->altitude) < altitudeError)
	;
}

boolean GPSReport::isLaterThan(GPSReport* that)
{
    return (this->date == that->date && this->time > that->time)
	|| (this->date > that->date);
}

void GPSReport::print()
{
    Serial.print((char*)this->id);
    Serial.print(' ');
    Serial.print(this->date);
    Serial.print(' ');
    Serial.print(this->time);
    Serial.print(' ');
    Serial.print(this->latitude);
    Serial.print(' ');
    Serial.print(this->longitude);
    Serial.print(' ');
    Serial.print(this->altitude);
    Serial.print(' ');
    Serial.print(this->speed);
    Serial.print(' ');
    Serial.print(this->course);
    Serial.print(' ');
    Serial.print(this->flags, HEX);
    Serial.print(' ');
    Serial.print(this->status, HEX);
    Serial.print(' ');
    Serial.println(this->hops);
}

GPSNet::GPSNet()
{
    _lastSecond = 0;
    _receivedReportsInLastSecond = 0;
    _receptionRate = 1.0; // Initial guess
    _updateTransmitInterval = 5000;
    _numReports = 0;
    _lastUpdatedReportIndex = 0;
    _transmitFunction = NULL;
    // Default constructor for GPSReport clears each entry in _reports to all 0s. DOES IT?
}

uint8_t GPSNet::numReports()
{
    return _numReports;
}

void GPSNet::scheduleReportForUpdate(uint8_t index)
{
    if (index == _lastUpdatedReportIndex)
	_lastUpdatedReportIndex = _reports[index].prev;
    unlinkReportWithIndex(index);
    linkReportWithIndexAfterIndex(index, _lastUpdatedReportIndex);
}

// New report about a node (may actually be us)
void GPSNet::newReport(GPSReport *report)
{
    uint8_t i;

    // See if its already there
    uint8_t index;
    GPSReport* existing = reportWithId(report->id, &index);
    if (existing)
    {
	if (report->isLaterThan(existing))
	{
	    // We have an earlier report for this ID, update it
	    *existing = *report; // Builtin copy
	    // Schedule this for the next update by linking from the last updated to this one
	    scheduleReportForUpdate(index);
	}
    }
    else
    {
	// Need to add a new report to our set (if there is room)
	if (_numReports < GPSNET_NUM_REPORTS)
	{
	    // There is room, allocate a new report and link it in to the last one
	    index = _numReports++;
	    if (index)
		linkReportWithIndexAfterIndex(index, index-1); // Not the first
	}
	else
	{
	    // Take over the LRU (pointed to by _lastUpdatedReport)
	    index = _lastUpdatedReportIndex;
	}
	existing = reportWithIndex(index);
	*existing = *report; // Builtin copy
	// Schedule this for the next update by linking from the last updated to this one
	scheduleReportForUpdate(index);
    }
//    printReports();
}

GPSReport* GPSNet::reportWithId(uint8_t* id, uint8_t *index)
{
    uint8_t i;
    // CAUTION: linear search
    for (i = 0; i < GPSNET_NUM_REPORTS; i++)
    {
	if (_reports[i].report.hasId(id))
	{
	    if (index) *index = i;
	    return &_reports[i].report;
	}
    }
    return NULL; // Not present
}

GPSReport* GPSNet::reportWithIndex(uint8_t index)
{
    if (index < GPSNET_NUM_REPORTS)
	return &_reports[index].report;
    return NULL; // Not present
}

void GPSNet::printReports()
{
    uint8_t i;

    for (i = 0; i < _numReports; i++)
    {
	Serial.print(i);
	Serial.print(' ');
	Serial.print(_reports[i].next);
	Serial.print(' ');
	Serial.print(_reports[i].prev);
	Serial.print(' ');
	_reports[i].report.print();
    }
}

// You can either override this or set a function to call
void GPSNet::transmitUpdate(GPSReport* report)
{
    if (_transmitFunction)
	(*_transmitFunction)(this, report);
}

void GPSNet::transmitNextUpdate()
{
    if (_numReports)
    {
	_lastUpdatedReportIndex = _reports[_lastUpdatedReportIndex].next;
	transmitUpdate(&_reports[_lastUpdatedReportIndex].report);
    }
}

// Report was received from another node
void GPSNet::receivedReport(GPSReport* report)
{
    _receivedReportsInLastSecond++;
    report->hops++;
    newReport(report);
}

// Based on the measured messsage reception rate, compute a sensible
// transmit rate.
// Call this once per second
void GPSNet::evaluateReceptionRate()
{
    // Update moving window average of reception rate
    _receptionRate += ((_receivedReportsInLastSecond - _receptionRate) / GPSNET_STATISTICS_WINDOW_SECS);
    _receivedReportsInLastSecond = 0;

    // Compute a time interval with some random jitter
    // REVISIT: should these parameters be configurable?
    _updateTransmitInterval = map(_receptionRate, 0, 5, 5000, 60000);
    _updateTransmitInterval += random(GPSNET_TRANSMIT_TIME_JITTER ); // Up to one second random jitter
}

void GPSNet::perSecondTasks()
{
    // One second housekeeping
    evaluateReceptionRate();
}

// Do various periodic tasks
void GPSNet::poll()
{
    uint32_t currentMillis = millis();
    if (currentMillis > _lastSecond + 1000)
    {
	perSecondTasks();
	_lastSecond = currentMillis;
    }

    // Time between transmissions is throttled by the reception rate
    // REVISIT: randomise times
    if (currentMillis > _lastUpdateTransmitTime + _updateTransmitInterval)
    {
	transmitNextUpdate();
	_lastUpdateTransmitTime = currentMillis;
    }
}

void GPSNet::unlinkReportWithIndex(uint8_t index)
{
    uint8_t next = _reports[index].next;
    uint8_t prev = _reports[index].prev;

    _reports[prev].next = next;
    _reports[next].prev = prev;
}

void GPSNet::linkReportWithIndexAfterIndex(uint8_t index, uint8_t after)
{
    uint8_t next = _reports[after].next;

    _reports[after].next = index;
    _reports[index].next = next;
    _reports[index].prev = after;
    _reports[next].prev = index;
}

void GPSNet::setTransmitFunction(void (*transmitFunction)(GPSNet* net, GPSReport* report))
{
    _transmitFunction = transmitFunction;
}
